import asyncio
from logging import getLogger
from typing import Dict, List, NamedTuple, Union

import aiohttp
import tabulate
from asyncache import cached
from cachetools import TTLCache
from redbot.core import VersionInfo, commands
from redbot.core import version_info as red_version
from redbot.core.utils.chat_formatting import box

from vexcogutils.loop import VexLoop

from .consts import CHECK, CROSS, DOCS_BASE
from .version import __version__ as utils_version

log = getLogger("red.vex-utils")


cog_ver_cache: TTLCache = TTLCache(maxsize=16, ttl=300)  # ttl is 5 mins
cog_ver_lock = asyncio.Lock()


def format_help(self: commands.Cog, ctx: commands.Context) -> str:
    """Wrapper for format_help_for_context. **Not** currently for use outside my cogs.

    Thanks Sinbad.

    Parameters
    ----------
    self : commands.Cog
        The Cog class
    context : commands.Context
        Context

    Returns
    -------
    str
        Formatted help
    """
    docs = DOCS_BASE.format(self.qualified_name.lower())
    pre_processed = super(type(self), self).format_help_for_context(ctx)

    return (
        f"{pre_processed}\n\nAuthor: **`{self.__author__}`**\nCog Version: "
        f"**`{self.__version__}`**\n{docs}"
    )
    # adding docs link here so doesn't show up in auto generated docs


# TODO: get utils version directly from pypi and stop using red internal util


async def format_info(
    qualified_name: str,
    cog_version: str,
    extras: Dict[str, Union[str, bool]] = {},
    loops: List[VexLoop] = [],
) -> str:
    """Generate simple info text about the cog. **Not** currently for use outside my cogs.

    Parameters
    ----------
    qualified_name : str
        The name you want to show, eg "BetterUptime"
    cog_version : str
        The version of the cog
    extras : Dict[str, Union[str, bool]], optional
        Dict which is foramtted as key: value\\n. Bools as a value will be replaced with
        check/cross emojis, by default {}
    loops : List[VexLoop], optional
        List of VexLoops you want to show, by default []

    Returns
    -------
    str
        Simple info text.
    """
    print(cog_ver_cache)
    try:
        latest = await _get_latest_vers()

        cog_updated = (
            CHECK
            if VersionInfo.from_str(cog_version)
            >= VersionInfo.from_str(latest.cogs.get(qualified_name.lower()))
            else CROSS
        )
        utils_updated = CHECK if VersionInfo.from_str(utils_version) >= latest.utils else CROSS
        red_updated = CHECK if red_version >= latest.red else CROSS
    except Exception:  # anything and everything, eg aiohttp error or version parsing error
        log.warning("Unable to parse versions.", exc_info=True)
        cog_updated = "Unknown"
        utils_updated = "Unknown"
        red_updated = "Unknown"

    start = f"{qualified_name} by Vexed.\n<https://github.com/Vexed01/Vex-Cogs>\n\n"
    versions = [
        ["Cog", cog_version, cog_updated],
        ["Utils", utils_version, utils_updated],
        ["Red", str(red_version), red_updated],
    ]

    data = []
    if loops:
        for loop in loops:
            data.append([loop.friendly_name, CHECK if loop.integrity else CROSS])

    if extras:
        if data:
            data.append([])
        for key, value in extras.items():
            if isinstance(value, bool):
                str_value = CHECK if value else CROSS
            else:
                assert isinstance(value, str)
                str_value = value
            data.append([key, str_value])

    boxed = box(tabulate.tabulate(versions, headers=["", "Version", "Up to date?"]))
    if data:
        boxed += box(tabulate.tabulate(data, tablefmt="plain"))

    return f"{start}{boxed}"


async def out_of_date_check(cogname: str, currentver: str) -> None:
    """Send a log at warning level if the cog is out of date."""
    try:
        async with cog_ver_lock:
            vers = await _get_latest_vers()
    except Exception as e:
        log.debug(
            f"Something went wrong checking if {cogname} cog is up to date. See below.", exc_info=e
        )
        # really doesn't matter if this fails so fine with debug level
        return
    if VersionInfo.from_str(currentver) < VersionInfo.from_str(vers.cogs.get(cogname)):
        log.warning(
            f"Your {cogname} cog, from Vex, is out of date. You can update your cogs with the "
            "'cog update' command in Discord."
        )
    else:
        log.debug(f"{cogname} cog is up to date")


class Vers(NamedTuple):
    cogs: Dict[str, str]
    utils: VersionInfo
    red: VersionInfo


@cached(cog_ver_cache)  # ttl is 5 mins
async def _get_latest_vers() -> Vers:
    data: dict
    print("hi!!")
    async with aiohttp.ClientSession() as session:
        async with session.get(
            "https://static.vexcodes.com/v1/versions.json", timeout=3  # ik its called static :)
        ) as r:
            data = await r.json()
            latest_cog = data.get("cogs", {})
        async with session.get("https://pypi.org/pypi/Red-DiscordBot/json", timeout=3) as r:
            data = await r.json()
            latest_red = VersionInfo.from_str(data.get("info", {}).get("version", "0.0.0"))
        async with session.get("https://pypi.org/pypi/vex-cog-utils/json", timeout=3) as r:
            data = await r.json()
            latest_utils = VersionInfo.from_str(data.get("info", {}).get("version", "0.0.0"))

    return Vers(latest_cog, latest_utils, latest_red)



DOCS_BASE = (
    "This cog has docs! Check them out at\nhttps://cogdocs.vexcodes.com/en/latest/cogs/{}.html"
)

CHECK = "\N{HEAVY CHECK MARK}\N{VARIATION SELECTOR-16}"
CROSS = "\N{CROSS MARK}"

SNOWFLAKE_REGEX = r"\b\d{17,20}\b"

SENTRY_MASTER_MSG = (
    "Hey there! This looks like the first time you're using Vexed's cogs (or you just updated to "
    "a version which supports this). To help make this cog, and all my others, as good "
    "and bug-free as possible, I have **opt-in** telemetry and error reporting __which affects "
    "all of my (github.com/Vexed01's) cogs__ on the Vex-Cogs repository, using Sentry. The "
    "telemetry consists of data on the cog release and performance data of backgroup tasks and "
    "loops (if applicable), and error reporting means that if something goes wrong the error and "
    "some associated data will be automatically sent to me so I can fix it quickly.\n\nA best "
    "effort is made to ensure no sensitive data is transmitted. For more information, including "
    "some technical details, visit <https://cogdocs.vexcodes.com/en/latest/telemetry.html>\n\n"
    "**If you would like to opt-in to telemetry and error reporting, and help me develop my cogs, "
    "run the command `[p]vextelemetry enable`. `[p]` is your prefix.**\nNo data is collected "
    "relating to command usage."
)

SENTRY_REMINDER_ON = (
    "Hey there! You just installed Vexed's {} cog. This is a reminder that you previously enabled "
    "telemetry and error reporting, which applies to all of my cogs, and this one is no "
    "different.\n\nI would like to emphasise again that a best effort it made to remove sensitive "
    "data. You can see <https://cogdocs.vexcodes.com/en/latest/telemetry.html> for more details "
    "and change your choice at any time with the `[p]vextelemetry` command, applying to all my "
    "cogs."
)

SENTRY_REMINDER_OFF = (
    "Hey there! You just installed Vexed's {} cog. This is a reminder that you previously chose "
    "not to enable telemetry and error reporting, which is also available in this cog. I hope you "
    "don't mind this reminder.\n\nI would like to emphasise again that a best effort it made to "
    "remove sensitive data. You can see <https://cogdocs.vexcodes.com/en/latest/telemetry.html> "
    "for more details and change your choice at any time with the `[p]vextelemetry` command, "
    "applying to all my cogs."
)

SENTRY_DSNS = {
    "wol": "https://ea2483bae296429797b23786fbf25980@o966432.ingest.sentry.io/5917446",
    "timechannel": "https://ae04709f0943421194b0d2c5f1467c85@o966432.ingest.sentry.io/5918790",
    "system": "https://0a9b0fa955ab4d2fb00b971ad823a3bf@o966432.ingest.sentry.io/5918773",
    "status": "https://f9eb2159800b4662b9dc266d7ee51c56@o966432.ingest.sentry.io/5918791",
    "stattrack": "https://dd7066711599488da627c8cc7de6ece3@o966432.ingest.sentry.io/5918793",
    "madtranslate": "https://405236d907214e6a9449e29ef1946535@o966432.ingest.sentry.io/5918774",
    "github": "https://9885bb05846f4adf966f746b0ebcc953@o966432.ingest.sentry.io/5918777",
    "ghissues": "https://7ecdcfcc80c04d08969b53c8e97df91c@o966432.ingest.sentry.io/5918779",
    "cmdlog": "https://4b39703d4d0041ceabc92bc554ddbc6e@o966432.ingest.sentry.io/5918794",
    "betteruptime": "https://5e84069469a24224ac03e290e51ea01e@o966432.ingest.sentry.io/5918797",
    "beautify": "https://52bcb80b8ecd485989ad105cdce1f1d4@o966432.ingest.sentry.io/5918798",
    "anotherpingcog": "https://8bef92be13f84d7dbdbb4e4ded63ad4c@o966432.ingest.sentry.io/5918803",
    "aliases": "https://0eda8fb326df45f58415975386a704c8@o966432.ingest.sentry.io/5918805",
}



from typing import TYPE_CHECKING, Optional

from redbot.core.bot import Red

from .chat import humanize_bytes, inline_hum_list
from .meta import format_help, format_info, out_of_date_check
from .sentry import SentryHelper as _SentryHelper
from .version import __version__

# if importlib.reload is ran, we don't want to re-initiate & overwrite SentryHelper - importlib
# keeps global variables
# to reiterate: SENTRY IS OPT-IN
if TYPE_CHECKING:
    sentryhelper = _SentryHelper()
    bot: Optional[Red] = None
else:
    try:
        sentryhelper
    except NameError:
        sentryhelper = _SentryHelper()
        bot = None
